# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Provides the top-level :ref:`convenience-api`, which allows you to create
many plots using a single compact statement.
"""

import logging
import sys

if sys.version_info.major < 3:
    raise RuntimeError("As of version 0.19, Toyplot requires Python 3.  For Python 2 support, use Toyplot version 0.18")

from toyplot.canvas import Canvas

__version__ = "1.0.2"

log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)
log.addHandler(logging.NullHandler())


class DeprecationWarning(Warning):
    """Used with :func:`warnings.warn` to mark deprecated API."""
    pass


def bars(
        a,
        b=None,
        c=None,
        along="x",
        baseline="stacked",
        color=None,
        filename=None,
        height=None,
        hyperlink=None,
        label=None,
        margin=50,
        opacity=1.0,
        padding=10,
        show=True,
        style=None,
        title=None,
        width=None,
        xlabel=None,
        xmax=None,
        xmin=None,
        xscale="linear",
        xshow=True,
        ylabel=None,
        ymax=None,
        ymin=None,
        yscale="linear",
        yshow=True,
    ):
    """Convenience function for creating a bar plot in a single call.

    See :meth:`toyplot.coordinates.Cartesian.bars`,
    :meth:`toyplot.canvas.Canvas.cartesian`, and :class:`toyplot.canvas.Canvas` for
    parameter descriptions.

    Returns
    -------
    canvas: :class:`toyplot.canvas.Canvas`
      A new canvas object.
    axes: :class:`toyplot.coordinates.Cartesian`
      A new set of 2D axes that fill the canvas.
    mark: :class:`toyplot.mark.BarMagnitudes` or :class:`toyplot.mark.BarBoundaries`
      The new bar mark.
    """
    canvas = Canvas(width=width, height=height)
    axes = canvas.cartesian(
        label=label,
        margin=margin,
        padding=padding,
        show=show,
        xlabel=xlabel,
        xmax=xmax,
        xmin=xmin,
        xscale=xscale,
        xshow=xshow,
        ylabel=ylabel,
        ymax=ymax,
        ymin=ymin,
        yscale=yscale,
        yshow=yshow,
        )

    mark = axes.bars(
        a=a,
        b=b,
        c=c,
        along=along,
        baseline=baseline,
        color=color,
        filename=filename,
        hyperlink=hyperlink,
        opacity=opacity,
        style=style,
        title=title,
        )
    return canvas, axes, mark


def fill(
        a,
        b=None,
        c=None,
        along="x",
        baseline=None,
        color=None,
        filename=None,
        height=None,
        label=None,
        margin=50,
        opacity=1.0,
        padding=10,
        show=True,
        style=None,
        title=None,
        width=None,
        xlabel=None,
        xmax=None,
        xmin=None,
        xscale="linear",
        xshow=True,
        ylabel=None,
        ymax=None,
        ymin=None,
        yscale="linear",
        yshow=True,
    ):
    """Convenience function for creating a fill plot in a single call.

    See :meth:`toyplot.coordinates.Cartesian.fill`,
    :meth:`toyplot.canvas.Canvas.cartesian`, and :class:`toyplot.canvas.Canvas` for
    parameter descriptions.

    Returns
    -------
    canvas: :class:`toyplot.canvas.Canvas`
      A new canvas object.
    axes: :class:`toyplot.coordinates.Cartesian`
      A new set of 2D axes that fill the canvas.
    mark: :class:`toyplot.mark.FillBoundaries` or :class:`toyplot.mark.FillMagnitudes`
      The new fill mark.
    """
    canvas = Canvas(
        height=height,
        width=width,
        )
    axes = canvas.cartesian(
        label=label,
        margin=margin,
        padding=padding,
        show=show,
        xlabel=xlabel,
        xmax=xmax,
        xmin=xmin,
        xscale=xscale,
        xshow=xshow,
        ylabel=ylabel,
        ymax=ymax,
        ymin=ymin,
        yscale=yscale,
        yshow=yshow,
        )
    mark = axes.fill(
        a=a,
        b=b,
        c=c,
        along=along,
        baseline=baseline,
        color=color,
        filename=filename,
        opacity=opacity,
        style=style,
        title=title,
        )
    return canvas, axes, mark


def graph(
        a,
        b=None,
        c=None,
        along="x",
        ecolor=None,
        eopacity=1.0,
        estyle=None,
        ewidth=1.0,
        height=None,
        hmarker=None,
        layout=None,
        margin=50,
        mmarker=None,
        mposition=0.5,
        olayout=None,
        padding=20,
        tmarker=None,
        varea=None,
        vcolor=None,
        vcoordinates=None,
        vlabel=None,
        vlshow=True,
        vlstyle=None,
        vmarker="o",
        vopacity=1.0,
        vsize=None,
        vstyle=None,
        vtitle=None,
        width=None,
    ): # pragma: no cover
    """Convenience function for creating a graph plot in a single call.

    See :meth:`toyplot.coordinates.Cartesian.graph`,
    :meth:`toyplot.canvas.Canvas.cartesian`, and :class:`toyplot.canvas.Canvas` for
    parameter descriptions.

    Returns
    -------
    canvas: :class:`toyplot.canvas.Canvas`
        A new canvas object.
    axes: :class:`toyplot.coordinates.Cartesian`
        A new set of 2D axes that fill the canvas.
    mark: :class:`toyplot.mark.Graph`
        The new graph mark.
    """
    canvas = Canvas(
        height=height,
        width=width,
        )
    axes = canvas.cartesian(
        aspect="fit-range",
        margin=margin,
        padding=padding,
        show=False,
        )
    mark = axes.graph(
        a=a,
        b=b,
        c=c,
        along=along,
        ecolor=ecolor,
        eopacity=eopacity,
        estyle=estyle,
        ewidth=ewidth,
        hmarker=hmarker,
        layout=layout,
        mmarker=mmarker,
        mposition=mposition,
        olayout=olayout,
        tmarker=tmarker,
        varea=varea,
        vcolor=vcolor,
        vcoordinates=vcoordinates,
        vlabel=vlabel,
        vlshow=vlshow,
        vlstyle=vlstyle,
        vmarker=vmarker,
        vopacity=vopacity,
        vsize=vsize,
        vstyle=vstyle,
        vtitle=vtitle,
        )
    return canvas, axes, mark


def image(
        data,
        height=None,
        margin=0,
        width=None,
    ): # pragma: no cover
    """Convenience function for displaying an image in a single call.

    See :meth:`toyplot.canvas.Canvas.image`, and :class:`toyplot.canvas.Canvas`
    for parameter descriptions.

    Returns
    -------
    canvas: :class:`toyplot.canvas.Canvas`
        A new canvas object.
    mark: :class:`toyplot.mark.Image`
        The new image mark.
    """
    canvas = Canvas(
        height=height,
        width=width,
        )
    mark = canvas.image(
        data=data,
        margin=margin,
        )
    return canvas, mark


def matrix(
        data,
        blabel=None,
        blocator=None,
        bshow=None,
        colorshow=False,
        filename=None,
        height=None,
        label=None,
        llabel=None,
        llocator=None,
        lshow=None,
        margin=50,
        rlabel=None,
        rlocator=None,
        rshow=None,
        step=1,
        tlabel=None,
        tlocator=None,
        tshow=None,
        width=None,
    ):
    """Convenience function to create a matrix visualization in a single call.

    See :meth:`toyplot.canvas.Canvas.matrix`, and
    :class:`toyplot.canvas.Canvas` for parameter descriptions.

    Returns
    -------
    canvas: :class:`toyplot.canvas.Canvas`
      A new canvas object.
    table: :class:`toyplot.coordinates.Table`
      A new set of table axes that fill the canvas.
    """
    canvas = Canvas(
        height=height,
        width=width,
        )
    axes = canvas.matrix(
        blabel=blabel,
        blocator=blocator,
        bshow=bshow,
        colorshow=colorshow,
        data=data,
        filename=filename,
        label=label,
        llabel=llabel,
        llocator=llocator,
        lshow=lshow,
        margin=margin,
        rlabel=rlabel,
        rlocator=rlocator,
        rshow=rshow,
        step=step,
        tlabel=tlabel,
        tlocator=tlocator,
        tshow=tshow,
        )
    return canvas, axes


def plot(
        a,
        b=None,
        along="x",
        area=None,
        aspect=None,
        color=None,
        filename=None,
        height=None,
        label=None,
        margin=50,
        marker=None,
        mfill=None,
        mlstyle=None,
        mopacity=1.0,
        mstyle=None,
        mtitle=None,
        opacity=1.0,
        padding=10,
        show=True,
        size=None,
        stroke_width=2.0,
        style=None,
        title=None,
        width=None,
        xlabel=None,
        xmax=None,
        xmin=None,
        xscale="linear",
        xshow=True,
        ylabel=None,
        ymax=None,
        ymin=None,
        yscale="linear",
        yshow=True,
    ):
    """Convenience function for creating a line plot in a single call.

    See :meth:`toyplot.coordinates.Cartesian.plot`,
    :meth:`toyplot.canvas.Canvas.cartesian`, and :class:`toyplot.canvas.Canvas` for
    parameter descriptions.

    Returns
    -------
    canvas: :class:`toyplot.canvas.Canvas`
      A new canvas object.
    axes: :class:`toyplot.coordinates.Cartesian`
      A new set of 2D axes that fill the canvas.
    mark: :class:`toyplot.mark.Plot`
      The new plot mark.
    """
    canvas = Canvas(
        height=height,
        width=width,
        )
    axes = canvas.cartesian(
        aspect=aspect,
        label=label,
        margin=margin,
        padding=padding,
        show=show,
        xlabel=xlabel,
        xmax=xmax,
        xmin=xmin,
        xscale=xscale,
        xshow=xshow,
        ylabel=ylabel,
        ymax=ymax,
        ymin=ymin,
        yscale=yscale,
        yshow=yshow,
        )
    mark = axes.plot(
        a=a,
        b=b,
        along=along,
        area=area,
        color=color,
        filename=filename,
        marker=marker,
        mfill=mfill,
        mlstyle=mlstyle,
        mopacity=mopacity,
        mstyle=mstyle,
        mtitle=mtitle,
        opacity=opacity,
        size=size,
        stroke_width=stroke_width,
        style=style,
        title=title,
        )
    return canvas, axes, mark


def scatterplot(
        a,
        b=None,
        along="x",
        area=None,
        aspect=None,
        color=None,
        filename=None,
        height=None,
        hyperlink=None,
        label=None,
        margin=50,
        marker="o",
        mlstyle=None,
        mstyle=None,
        opacity=1.0,
        padding=10,
        show=True,
        size=None,
        title=None,
        width=None,
        xlabel=None,
        xmax=None,
        xmin=None,
        xscale="linear",
        xshow=True,
        ylabel=None,
        ymax=None,
        ymin=None,
        yscale="linear",
        yshow=True,
    ):
    """Convenience function for creating a scatter plot in a single call.

    See :meth:`toyplot.coordinates.Cartesian.scatterplot`,
    :meth:`toyplot.canvas.Canvas.cartesian`, and :class:`toyplot.canvas.Canvas` for
    parameter descriptions.

    Returns
    -------
    canvas: :class:`toyplot.canvas.Canvas`
      A new canvas object.
    axes: :class:`toyplot.coordinates.Cartesian`
      A new set of 2D axes that fill the canvas.
    mark: :class:`toyplot.mark.Point`
      The new scatterplot mark.
    """
    canvas = Canvas(
        height=height,
        width=width,
        )
    axes = canvas.cartesian(
        aspect=aspect,
        label=label,
        margin=margin,
        padding=padding,
        show=show,
        xlabel=xlabel,
        xmax=xmax,
        xmin=xmin,
        xscale=xscale,
        xshow=xshow,
        ylabel=ylabel,
        ymax=ymax,
        ymin=ymin,
        yscale=yscale,
        yshow=yshow,
        )
    mark = axes.scatterplot(
        a=a,
        b=b,
        along=along,
        area=area,
        color=color,
        filename=filename,
        hyperlink=hyperlink,
        marker=marker,
        mlstyle=mlstyle,
        mstyle=mstyle,
        opacity=opacity,
        size=size,
        title=title,
        )
    return canvas, axes, mark


def table(
        data=None,
        brows=None,
        columns=None,
        filename=None,
        height=None,
        label=None,
        lcolumns=None,
        margin=50,
        rcolumns=None,
        rows=None,
        trows=None,
        width=None,
    ):
    """Convenience function to create a table visualization in a single call.

    See :meth:`toyplot.canvas.Canvas.table`, and :class:`toyplot.canvas.Canvas`
    for parameter descriptions.

    Returns
    -------
    canvas: :class:`toyplot.canvas.Canvas`
      A new canvas object.
    table: :class:`toyplot.coordinates.Table`
      A new set of table axes that fill the canvas.
    """
    canvas = Canvas(
        height=height,
        width=width,
        )
    axes = canvas.table(
        brows=brows,
        columns=columns,
        data=data,
        filename=filename,
        label=label,
        lcolumns=lcolumns,
        margin=margin,
        rcolumns=rcolumns,
        rows=rows,
        trows=trows,
        )
    return canvas, axes



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Support functions for rendering using ReportLab."""


import base64
import io
import re

import numpy
import reportlab.lib.colors
import reportlab.lib.utils

import toyplot.color
import toyplot.units


def render(svg, canvas):
    """Render the SVG representation of a toyplot canvas to a ReportLab canvas.

    Parameters
    ----------
    svg: xml.etree.ElementTree.Element
      SVG representation of a :class:`toyplot.canvas.Canvas` returned by
      :func:`toyplot.svg.render()`.

    canvas: reportlab.pdfgen.canvas.Canvas
      ReportLab canvas that will be used to render the plot.
    """
    def get_fill(root, style):
        if "fill" not in style:
            return None, None # pragma: no cover

        gradient_id = re.match("^url[(]#(.*)[)]$", style["fill"])
        if gradient_id:
            gradient_id = gradient_id.group(1)
            gradient_xml = root.find(".//*[@id='%s']" % gradient_id)
            if gradient_xml.tag != "linearGradient":
                raise NotImplementedError("Only linear gradients are implemented.") # pragma: no cover
            if gradient_xml.get("gradientUnits") != "userSpaceOnUse":
                raise NotImplementedError("Only userSpaceOnUse gradients are implemented.") # pragma: no cover
            return None, gradient_xml

        color = toyplot.color.css(style["fill"])
        if color is None:
            return None, None

        fill_opacity = float(style.get("fill-opacity", 1.0))
        opacity = float(style.get("opacity", 1.0))
        fill = toyplot.color.rgba(
            color["r"],
            color["g"],
            color["b"],
            color["a"] * fill_opacity * opacity,
            )
        return fill, None

    def get_stroke(style):
        if "stroke" not in style:
            return None # pragma: no cover

        color = toyplot.color.css(style["stroke"])
        if color is None:
            return None

        stroke_opacity = float(style.get("stroke-opacity", 1.0))
        opacity = float(style.get("opacity", 1.0))
        return toyplot.color.rgba(
            color["r"],
            color["g"],
            color["b"],
            color["a"] * stroke_opacity * opacity,
            )

    def get_line_cap(style):
        if "stroke-linecap" not in style:
            return 0
        elif style["stroke-linecap"] == "butt":
            return 0
        elif style["stroke-linecap"] == "round":
            return 1
        elif style["stroke-linecap"] == "square":
            return 2

    def get_font_family(style):
        if "font-family" not in style:
            return None # pragma: no cover

        bold = True if style.get("font-weight", "") == "bold" else False
        italic = True if style.get("font-style", "") == "italic" else False
        for font_family in style["font-family"].split(","):
            font_family = font_family.lower()
            if font_family in get_font_family.substitutions:
                font_family = get_font_family.substitutions[font_family]
                return get_font_family.font_table[(font_family, bold, italic)]

        raise ValueError("Unknown font family: %s" % style["font-family"]) # pragma: no cover

    get_font_family.font_table = {
        ("courier", False, False): "Courier",
        ("courier", True, False): "Courier-Bold",
        ("courier", False, True): "Courier-Oblique",
        ("courier", True, True): "Courier-BoldOblique",
        ("helvetica", False, False): "Helvetica",
        ("helvetica", True, False): "Helvetica-Bold",
        ("helvetica", False, True): "Helvetica-Oblique",
        ("helvetica", True, True): "Helvetica-BoldOblique",
        ("times", False, False): "Times-Roman",
        ("times", True, False): "Times-Bold",
        ("times", False, True): "Times-Italic",
        ("times", True, True): "Times-BoldItalic",
        }

    get_font_family.substitutions = {
        "courier": "courier",
        "helvetica": "helvetica",
        "monospace": "courier",
        "sans-serif": "helvetica",
        "serif": "times",
        "times": "times",
        }

    def set_fill_color(canvas, color):
        canvas.setFillColorRGB(color["r"], color["g"], color["b"])
        canvas.setFillAlpha(color["a"].item())

    def set_stroke_color(canvas, color):
        canvas.setStrokeColorRGB(color["r"], color["g"], color["b"])
        canvas.setStrokeAlpha(color["a"].item())

    def render_element(root, element, canvas, styles):
        canvas.saveState()

        current_style = {}
        if styles:
            current_style.update(styles[-1])
        for declaration in element.get("style", "").split(";"):
            if declaration == "":
                continue
            key, value = declaration.split(":")
            current_style[key] = value
        styles.append(current_style)

        if "stroke-width" in current_style:
            canvas.setLineWidth(float(current_style["stroke-width"]))

        if "stroke-dasharray" in current_style:
            canvas.setDash([float(length) for length in current_style["stroke-dasharray"].split(",")])

        if current_style.get("visibility") != "hidden":

            if "transform" in element.attrib:
                for transformation in element.get("transform").split(")")[::1]:
                    if transformation:
                        transform, arguments = transformation.split("(")
                        arguments = arguments.split(",")
                        if transform.strip() == "translate":
                            if len(arguments) == 2:
                                canvas.translate(float(arguments[0]), float(arguments[1]))
                        elif transform.strip() == "rotate":
                            if len(arguments) == 1:
                                canvas.rotate(float(arguments[0]))
                            if len(arguments) == 3:
                                canvas.translate(float(arguments[1]), float(arguments[2]))
                                canvas.rotate(float(arguments[0]))
                                canvas.translate(-float(arguments[1]), -float(arguments[2]))

            if element.tag == "svg":
                if "background-color" in current_style:
                    set_fill_color(canvas, toyplot.color.css(current_style["background-color"]))
                    canvas.rect(
                        0,
                        0,
                        float(element.get("width")[:-2]),
                        float(element.get("height")[:-2]),
                        stroke=0,
                        fill=1,
                        )

                if current_style["border-style"] != "none":
                    set_stroke_color(canvas, toyplot.color.css(current_style["border-color"]))
                    canvas.setLineWidth(float(current_style["border-width"]))
                    canvas.rect(
                        0,
                        0,
                        float(element.get("width")[:-2]),
                        float(element.get("height")[:-2]),
                        stroke=1,
                        fill=0,
                        )

                for child in element:
                    render_element(root, child, canvas, styles)

            elif element.tag == "a":
                # At the moment, it doesn't look like reportlab supports external hyperlinks.
                for child in element:
                    render_element(root, child, canvas, styles)

            elif element.tag == "g":
                if element.get("clip-path", None) is not None:
                    clip_id = element.get("clip-path")[5:-1]
                    clip_path = root.find(".//*[@id='%s']" % clip_id)
                    for child in clip_path:
                        if child.tag == "rect":
                            x = float(child.get("x"))
                            y = float(child.get("y"))
                            width = float(child.get("width"))
                            height = float(child.get("height"))
                            path = canvas.beginPath()
                            path.moveTo(x, y)
                            path.lineTo(x + width, y)
                            path.lineTo(x + width, y + height)
                            path.lineTo(x, y + height)
                            path.close()
                            canvas.clipPath(path, stroke=0, fill=1)
                        else:
                            toyplot.log.error("Unhandled clip tag: %s", child.tag) # pragma: no cover

                for child in element:
                    render_element(root, child, canvas, styles)

            elif element.tag == "clipPath":
                pass

            elif element.tag == "line":
                stroke = get_stroke(current_style)
                if stroke is not None:
                    set_stroke_color(canvas, stroke)
                    canvas.setLineCap(get_line_cap(current_style))
                    canvas.line(
                        float(element.get("x1", 0)),
                        float(element.get("y1", 0)),
                        float(element.get("x2", 0)),
                        float(element.get("y2", 0)),
                        )
            elif element.tag == "path":
                stroke = get_stroke(current_style)
                if stroke is not None:
                    set_stroke_color(canvas, stroke)
                    canvas.setLineCap(get_line_cap(current_style))
                    path = canvas.beginPath()
                    commands = element.get("d").split()
                    while commands:
                        command = commands.pop(0)
                        if command == "L":
                            path.lineTo(
                                float(commands.pop(0)), float(commands.pop(0)))
                        elif command == "M":
                            path.moveTo(
                                float(commands.pop(0)), float(commands.pop(0)))
                    canvas.drawPath(path)
            elif element.tag == "polygon":
                fill, fill_gradient = get_fill(root, current_style)
                if fill_gradient is not None:
                    raise NotImplementedError("Gradient <polygon> not implemented.") # pragma: no cover
                if fill is not None:
                    set_fill_color(canvas, fill)
                stroke = get_stroke(current_style)
                if stroke is not None:
                    set_stroke_color(canvas, stroke)

                points = [point.split(",") for point in element.get("points").split()]
                path = canvas.beginPath()
                for point in points[:1]:
                    path.moveTo(float(point[0]), float(point[1]))
                for point in points[1:]:
                    path.lineTo(float(point[0]), float(point[1]))
                path.close()
                canvas.drawPath(path, stroke=stroke is not None, fill=fill is not None)
            elif element.tag == "rect":
                fill, fill_gradient = get_fill(root, current_style)
                if fill is not None:
                    set_fill_color(canvas, fill)
                stroke = get_stroke(current_style)
                if stroke is not None:
                    set_stroke_color(canvas, stroke)

                x = float(element.get("x", 0))
                y = float(element.get("y", 0))
                width = float(element.get("width"))
                height = float(element.get("height"))

                path = canvas.beginPath()
                path.moveTo(x, y)
                path.lineTo(x + width, y)
                path.lineTo(x + width, y + height)
                path.lineTo(x, y + height)
                path.close()

                if fill_gradient is not None:
                    pdf_colors = []
                    pdf_offsets = []
                    for stop in fill_gradient:
                        offset = float(stop.get("offset"))
                        color = toyplot.color.css(stop.get("stop-color"))
                        opacity = float(stop.get("stop-opacity"))
                        pdf_colors.append(reportlab.lib.colors.Color(color["r"], color["g"], color["b"], color["a"] * opacity))
                        pdf_offsets.append(offset)
                    canvas.saveState()
                    canvas.clipPath(path, stroke=0, fill=1)
                    canvas.setFillAlpha(1)
                    canvas.linearGradient(
                        float(fill_gradient.get("x1")),
                        float(fill_gradient.get("y1")),
                        float(fill_gradient.get("x2")),
                        float(fill_gradient.get("y2")),
                        pdf_colors,
                        pdf_offsets,
                        )
                    canvas.restoreState()

                canvas.drawPath(path, stroke=stroke is not None, fill=fill is not None)
            elif element.tag == "circle":
                fill, fill_gradient = get_fill(root, current_style)
                if fill_gradient is not None:
                    raise NotImplementedError("Gradient <circle> not implemented.") # pragma: no cover
                if fill is not None:
                    set_fill_color(canvas, fill)
                stroke = get_stroke(current_style)
                if stroke is not None:
                    set_stroke_color(canvas, stroke)

                cx = float(element.get("cx", 0))
                cy = float(element.get("cy", 0))
                r = float(element.get("r"))
                canvas.circle(cx, cy, r, stroke=stroke is not None, fill=fill is not None)
            elif element.tag == "text":
                x = float(element.get("x", 0))
                y = float(element.get("y", 0))
                fill, fill_gradient = get_fill(element, current_style)
                stroke = get_stroke(current_style)
                font_family = get_font_family(current_style)
                font_size = toyplot.units.convert(current_style["font-size"], target="px")
                text = element.text

                canvas.saveState()
                canvas.setFont(font_family, font_size)
                if fill is not None:
                    set_fill_color(canvas, fill)
                if stroke is not None:
                    set_stroke_color(canvas, stroke)
                canvas.translate(x, y)
                canvas.scale(1, -1)
                canvas.drawString(0, 0, text)
                canvas.restoreState()

            elif element.tag == "image":
                import PIL.Image
                image = element.get("xlink:href")
                if not image.startswith("data:image/png;base64,"):
                    raise ValueError("Unsupported image type.") # pragma: no cover
                image = base64.standard_b64decode(image[22:])
                image = io.BytesIO(image)
                image = PIL.Image.open(image)
                image = reportlab.lib.utils.ImageReader(image)

                x = float(element.get("x", 0))
                y = float(element.get("y", 0))
                width = float(element.get("width"))
                height = float(element.get("height"))

                canvas.saveState()
                path = canvas.beginPath()
                set_fill_color(canvas, toyplot.color.rgb(1, 1, 1))
                canvas.rect(x, y, width, height, stroke=0, fill=1)
                canvas.translate(x, y + height)
                canvas.scale(1, -1)
                canvas.drawImage(image=image, x=0, y=0, width=width, height=height, mask=None)
                canvas.restoreState()

            elif element.tag in ["defs", "title"]:
                pass

            else:
                raise Exception("unhandled tag: %s" % element.tag) # pragma: no cover

        styles.pop()
        canvas.restoreState()

    render_element(svg, svg, canvas, [])



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Functions for broadcasting values into 1D or 2D arrays.

The functions in this module are used by Toyplot to handle the conversion from
`constant`, `per-series`, and `per-datum` values into their canonical 1D and 2D
array representations.
"""


import numpy


def scalar(value, shape):
    """Return a 1D or 2D array of floats with the given shape.

    Parameters
    ----------
    shape: (tuple)
        Shape of the desired output array.  A 1-tuple will produce a 1D output
        vector containing :math:`N` per-series values.  A 2-tuple will produce an :math:`M \\times N`
        output matrix of per-datum values grouped into :math:`N` series.

    Returns
    -------
    array: :class:`numpy.ndarray`
    """
    array = numpy.array(value).astype("float64")
    # As a special-case, allow a vector with shape M to be matched-up with an
    # M x 1 matrix.
    if array.ndim == 1 and isinstance(shape, tuple) and len(
            shape) == 2 and array.shape[0] == shape[0] and shape[1] == 1:
        return numpy.reshape(array, shape)
    return numpy.broadcast_arrays(array, numpy.empty(shape))[0]


#def string(value, shape):
#    array = numpy.array(value).astype("unicode")
#    # As a special-case, allow a vector with shape M to be matched-up with an
#    # M x 1 matrix.
#    if array.ndim == 1 and isinstance(shape, tuple) and len(
#            shape) == 2 and array.shape[0] == shape[0] and shape[1] == 1:
#        return numpy.reshape(array, shape)
#    return numpy.broadcast_arrays(array, numpy.empty(shape))[0]


def pyobject(value, shape):
    """Return a 1D or 2D array of objects with the given shape.

    Parameters
    ----------
    shape: (tuple)
        Shape of the desired output array.  A 1-tuple will produce a 1D output
        vector containing :math:`N` per-series values.  A 2-tuple will produce an :math:`M \\times N`
        output matrix of per-datum values grouped into :math:`N` series.

    Returns
    -------
    array: :class:`numpy.ndarray`
    """
    array = numpy.array(value)
    # As a special-case, allow a vector with shape M to be matched-up with an
    # M x 1 matrix.
    if array.ndim == 1 and isinstance(shape, tuple) and len(
            shape) == 2 and array.shape[0] == shape[0] and shape[1] == 1:
        return numpy.reshape(array, shape)
    result = numpy.empty(shape, dtype="object")
    result.flat = [u for u, _ in numpy.broadcast(array, result)]
    return result



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

from setuptools import setup, find_packages
import re

setup(
    author="Timothy M. Shead",
    author_email="tshead@sandia.gov",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: Other Environment",
        "Environment :: Web Environment",
        "Framework :: IPython",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
    description="A modern plotting toolkit supporting electronic publishing and reproducibility.",
    install_requires=[
        "arrow>=1.0",
        "custom_inherit",
        "multipledispatch",
        "numpy>=1.8.0",
        "pypng",
        "reportlab",
    ],
    long_description="""Toyplot is the kid-sized plotting toolkit for Python with grownup-sized goals:
  * Develop beautiful plots that embrace the unique capabilities of electronic publishing and support repoducibility.
  * Create the best possible data graphics "out-of-the-box", maximizing data ink and minimizing chartjunk.
  * Provide a clean, minimalist interface that scientists and engineers will love.

  See the Toyplot documentation at http://toyplot.readthedocs.io, and the Toyplot sources at http://github.com/sandialabs/toyplot""",
    maintainer="Timothy M. Shead",
    maintainer_email="tshead@sandia.gov",
    name="toyplot",
    packages=find_packages(),
    package_data = {"": ["*.csv"]},
    project_urls={
        "Chat": "https://github.com/sandialabs/toyplot/discussions",
        "Coverage": "https://coveralls.io/r/sandialabs/toyplot",
        "Documentation": "https://toyplot.readthedocs.io",
        "Issue Tracker": "https://github.com/sandialabs/toyplot/issues",
        "Regression Tests": "https://github.com/sandialabs/toyplot/actions",
        "Source": "https://github.com/sandialabs/toyplot",
    },
    url="http://toyplot.readthedocs.io",
    version=re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        open(
            "toyplot/__init__.py",
            "r").read(),
        re.M).group(1),
)



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Generate MPEG-4 videos."""


import logging
import os
import subprocess

import toyplot.png

log = logging.getLogger(__name__)

# Verify that ffmpeg is installed, and check the version
_ffmpeg_command = None
_ffmpeg_version = None
for command in ["ffmpeg"]:
    try:
        _ffmpeg_version = subprocess.check_output([command, "-version"]).decode(encoding="utf-8").strip()
        _ffmpeg_command = command
        log.info("Using %s.", _ffmpeg_version)
    except: # pragma: no cover
        pass

def render(
        canvas,
        filename,
        width=None,
        height=None,
        scale=None,
        progress=None):
    """Render a canvas as an MPEG-4 video.

    By default, the canvas dimensions in CSS pixels are mapped directly to
    pixels in the output MPEG-4 video.  Use one of `width`, `height`, or
    `scale` to override this behavior.

    Parameters
    ----------
    canvas: :class:`toyplot.canvas.Canvas`
      Canvas to be rendered.
    filename: string
      Output video filename.
    width: number, optional
      Specify the width of the output video in pixels.
    height: number, optional
      Specify the height of the output video in pixels.
    scale: number, optional
      Ratio of output video pixels to `canvas` drawing units.
    progress: callback function taking a single `frame` argument, optional
      Callback function that will receive the number of each frame as it's
      written; useful to provide an indication of progress to end-users.

    Notes
    -----
    The individual video frames are rendered using PNG representations
    of the canvas generated with :func:`toyplot.png.render_frames()`.

    Examples
    --------
    >>> def callback(frame):
    ...   print "Writing frame %s" % frame
    ... toyplot.mp4.render(canvas, "test.mp4", progress=callback)
    """

    if _ffmpeg_command is None:
        raise RuntimeError("An ffmpeg executable is required.")  # pragma: no cover


    command = [
        _ffmpeg_command,
        "-f", "image2pipe",
        "-c", "png",
        "-i", "-",
        "-pix_fmt", "yuv420p",
        "-vcodec", "h264",
        "-preset", "slow",
        "-tune", "animation",
        "-crf", "17",
        "-y",
        filename,
    ]

    try:
        log.info("Running command: %s", " ".join(command))
        ffmpeg = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        for frame, png in enumerate(
                toyplot.png.render_frames(
                    canvas=canvas, width=width, height=height, scale=scale)):
            if progress is not None:
                progress(frame)
            ffmpeg.stdin.write(png)
        ffmpeg.stdin.close()
        ffmpeg.wait()
    except Exception as e: # pragma: no cover
        log.error(ffmpeg.stdout.read())
        log.error(ffmpeg.stderr.read())
        raise e
